	"FORMAT",	S_FORMAT,
	"FMT",	S_FMT,
	"LABEL",	S_LABEL,
	"LEFTSTRING",	S_LEFTSTRING,
	"RIGHTSTRING",	S_RIGHTSTRING,
	"GET",	S_GET,
	"PUT",	S_PUT,
	"MERGE",	S_MERGE,
	"LET",	S_LET,
	"WRITE",	S_WRITE,
	"TBL",	S_TBL,
	"COPY",	S_COPY,
	"SHOW",	S_SHOW,
	"ERASE",	S_ERASE,
	"FILL",	S_FILL,
	"LOCK",	S_LOCK,
	"UNLOCK",	S_UNLOCK,
	"GOTO",	S_GOTO,
	"DEFINE",	S_DEFINE,
	"UNDEFINE",	S_UNDEFINE,
	"VALUE",	S_VALUE,
	"MDIR",	S_MDIR,
	"HIDE",	S_HIDE,
	"SET",	S_SET,
