
typedef union  {
    int ival;
    double fval;
    struct ent_ptr ent;
    struct enode *enode;
    char *sval;
    struct range_s rval;
} YYSTYPE;
extern YYSTYPE yylval;
# define STRING 257
# define NUMBER 258
# define FNUMBER 259
# define RANGE 260
# define VAR 261
# define WORD 262
# define COL 263
# define S_FORMAT 264
# define S_FMT 265
# define S_LABEL 266
# define S_LEFTSTRING 267
# define S_RIGHTSTRING 268
# define S_GET 269
# define S_PUT 270
# define S_MERGE 271
# define S_LET 272
# define S_WRITE 273
# define S_TBL 274
# define S_COPY 275
# define S_SHOW 276
# define S_ERASE 277
# define S_FILL 278
# define S_LOCK 279
# define S_UNLOCK 280
# define S_GOTO 281
# define S_DEFINE 282
# define S_UNDEFINE 283
# define S_VALUE 284
# define S_MDIR 285
# define S_HIDE 286
# define S_SET 287
# define K_ERROR 288
# define K_INVALID 289
# define K_FIXED 290
# define K_SUM 291
# define K_PROD 292
# define K_AVG 293
# define K_STDDEV 294
# define K_COUNT 295
# define K_ABS 296
# define K_ACOS 297
# define K_ASIN 298
# define K_ATAN 299
# define K_ATAN2 300
# define K_CEIL 301
# define K_COS 302
# define K_EXP 303
# define K_FABS 304
# define K_FLOOR 305
# define K_HYPOT 306
# define K_LN 307
# define K_LOG 308
# define K_PI 309
# define K_POW 310
# define K_SIN 311
# define K_SQRT 312
# define K_TAN 313
# define K_DTR 314
# define K_RTD 315
# define K_MAX 316
# define K_MIN 317
# define K_RND 318
# define K_ROUND 319
# define K_IF 320
# define K_PV 321
# define K_FV 322
# define K_PMT 323
# define K_HOUR 324
# define K_MINUTE 325
# define K_SECOND 326
# define K_MONTH 327
# define K_DAY 328
# define K_YEAR 329
# define K_NOW 330
# define K_DATE 331
# define K_DTS 332
# define K_TTS 333
# define K_FMT 334
# define K_SUBSTR 335
# define K_UPPER 336
# define K_LOWER 337
# define K_CAPITAL 338
# define K_STON 339
# define K_EQS 340
# define K_EXT 341
# define K_NVAL 342
# define K_SVAL 343
# define K_LOOKUP 344
# define K_HLOOKUP 345
# define K_VLOOKUP 346
# define K_INDEX 347
# define K_STINDEX 348
# define K_AUTO 349
# define K_AUTOCALC 350
# define K_BYROWS 351
# define K_BYCOLS 352
# define K_ITERATIONS 353
# define K_NUMERIC 354
# define K_PRESCALE 355
# define K_EXTFUN 356
# define K_CELLCUR 357
# define K_TOPROW 358
# define K_TBLSTYLE 359
# define K_TBL 360
# define K_LATEX 361
# define K_SLATEX 362
# define K_TEX 363
# define K_FRAME 364
# define K_RNDINFINITY 365
# define K_MYROW 366
# define K_MYCOL 367
# define K_COLTOA 368
# define K_CRACTION 369
# define K_CRROW 370
# define K_CRCOL 371
# define K_ROWLIMIT 372
# define K_COLLIMIT 373
# define K_NUMITER 374
# define K_MATRIX_ADD 375
# define K_MATRIX_SUB 376
# define K_MATRIX_INV 377
# define K_MATRIX_MULT 378
# define K_MATRIX_TRANS 379
# define K_GRAPH 380
