	"ERROR",	K_ERROR,
	"INVALID",	K_INVALID,
	"FIXED",	K_FIXED,
	"SUM",	K_SUM,
	"PROD",	K_PROD,
	"AVG",	K_AVG,
	"STDDEV",	K_STDDEV,
	"COUNT",	K_COUNT,
	"ABS",	K_ABS,
	"ACOS",	K_ACOS,
	"ASIN",	K_ASIN,
	"ATAN",	K_ATAN,
	"ATAN2",	K_ATAN2,
	"CEIL",	K_CEIL,
	"COS",	K_COS,
	"EXP",	K_EXP,
	"FABS",	K_FABS,
	"FLOOR",	K_FLOOR,
	"HYPOT",	K_HYPOT,
	"LN",	K_LN,
	"LOG",	K_LOG,
	"PI",	K_PI,
	"POW",	K_POW,
	"SIN",	K_SIN,
	"SQRT",	K_SQRT,
	"TAN",	K_TAN,
	"DTR",	K_DTR,
	"RTD",	K_RTD,
	"MAX",	K_MAX,
	"MIN",	K_MIN,
	"RND",	K_RND,
	"ROUND",	K_ROUND,
	"IF",	K_IF,
	"PV",	K_PV,
	"FV",	K_FV,
	"PMT",	K_PMT,
	"HOUR",	K_HOUR,
	"MINUTE",	K_MINUTE,
	"SECOND",	K_SECOND,
	"MONTH",	K_MONTH,
	"DAY",	K_DAY,
	"YEAR",	K_YEAR,
	"NOW",	K_NOW,
	"DATE",	K_DATE,
	"DTS",	K_DTS,
	"TTS",	K_TTS,
	"FMT",	K_FMT,
	"SUBSTR",	K_SUBSTR,
	"UPPER",	K_UPPER,
	"LOWER",	K_LOWER,
	"CAPITAL",	K_CAPITAL,
	"STON",	K_STON,
	"EQS",	K_EQS,
	"EXT",	K_EXT,
	"NVAL",	K_NVAL,
	"SVAL",	K_SVAL,
	"LOOKUP",	K_LOOKUP,
	"HLOOKUP",	K_HLOOKUP,
	"VLOOKUP",	K_VLOOKUP,
	"INDEX",	K_INDEX,
	"STINDEX",	K_STINDEX,
	"AUTO",	K_AUTO,
	"AUTOCALC",	K_AUTOCALC,
	"BYROWS",	K_BYROWS,
	"BYCOLS",	K_BYCOLS,
	"ITERATIONS",	K_ITERATIONS,
	"NUMERIC",	K_NUMERIC,
	"PRESCALE",	K_PRESCALE,
	"EXTFUN",	K_EXTFUN,
	"CELLCUR",	K_CELLCUR,
	"TOPROW",	K_TOPROW,
	"TBLSTYLE",	K_TBLSTYLE,
	"TBL",	K_TBL,
	"LATEX",	K_LATEX,
	"SLATEX",	K_SLATEX,
	"TEX",	K_TEX,
	"FRAME",	K_FRAME,
	"RNDINFINITY",	K_RNDINFINITY,
	"MYROW",	K_MYROW,
	"MYCOL",	K_MYCOL,
	"COLTOA",	K_COLTOA,
	"CRACTION",	K_CRACTION,
	"CRROW",	K_CRROW,
	"CRCOL",	K_CRCOL,
	"ROWLIMIT",	K_ROWLIMIT,
	"COLLIMIT",	K_COLLIMIT,
	"NUMITER",	K_NUMITER,
	"MATRIX_ADD",	K_MATRIX_ADD,
	"MATRIX_SUB",	K_MATRIX_SUB,
	"MATRIX_INV",	K_MATRIX_INV,
	"MATRIX_MULT",	K_MATRIX_MULT,
	"MATRIX_TRANS",	K_MATRIX_TRANS,
	"GRAPH",	K_GRAPH,
